! ******************************************************* !
!
subroutine calc_alpha_correction
!
! ******************************************************* !
!  
!     computing alpha correction used in propagators
!     from halo-fit
  use data,only : pi, ik_max, ak, ibox, Lbox, zred, &
       & ak_nl, rknl, ik_max_nl, etaa, alphaz, iz_max, izmax
  implicit none
  
  integer, parameter :: ixmax = 500
  integer  ik, iz, ix
  double precision z, zmax, denom, numer, growth
  double precision k, kmin, kmax, pklin, pknl
  double precision wk(ixmax), kk(ixmax)
  external growth
! -------------------------------------------------

  zmax = zred + 10.d0
  
  do iz = 1, izmax
     
     z = zred + (zmax-zred)*dble(iz-1)/dble(izmax-1)
     etaa(iz) = log(growth(z))
     
     call calc_halo_pk(z)

!     ////////  Compute alpha^2(z) correction  ////////  

     denom = 0.d0
     numer = 0.d0
     kmin = max( ak(1), ak_nl(1) )
     
     if(ibox == 0) kmin = 2.d0 * pi / Lbox
     
     kmax = min( ak(ik_max), ak_nl(ik_max_nl) ) 
     
     call gauleg(log(kmin), log(min(1.d0*rknl, kmax)), kk, wk, ixmax)
     
     do ix = 1, ixmax
        k = exp(kk(ix))
        call find_pk(k, pklin)
        call find_pknl(k, pknl)
        numer = numer + wk(ix) * pknl * k
        denom = denom + wk(ix) * pklin * k
     enddo
     
     alphaz(iz) = sqrt(numer / denom) / growth(z)
     
     if( alphaz(iz) <= 1.d0 ) cycle     
  enddo
  
  iz_max = iz - 1 
  return
end subroutine calc_alpha_correction
! ******************************************************* !
!
subroutine find_alpha(eta1, alpha_z)
!
! ******************************************************* !
  use data,only : etaa, alphaz, izmax, iz_max
  implicit none

  double precision, intent(in) :: eta1
  double precision, intent(out) :: alpha_z
  integer j, jmin, jmax
  double precision s, ds
! -------------------------------------------

  if( (eta1 <= etaa(1)) .and. (eta1 >= etaa(iz_max))) then

     call hunt(etaa, iz_max, eta1, j)
     
     jmin = max(1, j - 2)
     jmax = min(iz_max, j + 2)
     
     call polint(etaa(jmin:jmax),alphaz(jmin:jmax),jmax-jmin+1,eta1,s,ds)
     alpha_z = s

  else
     alpha_z = 1.d0
  endif
  return
end subroutine find_alpha
! ******************************************************* !
!
subroutine find_pknl(kk, pknl)
!
! ******************************************************* !
  use data,only : ak_nl, pk_nl, ik_max
  implicit none

  double precision, intent(in) :: kk
  double precision, intent(out) :: pknl
  integer j, jmin, jmax
  double precision s, ds
! -------------------------------------------

  call hunt(ak_nl, ik_max, kk, j)

  jmin = max(1, j - 2)
  jmax = min(ik_max, j + 2)

  call polint(ak_nl(jmin:jmax),pk_nl(jmin:jmax),jmax-jmin+1,kk,s,ds)
  pknl = s
  return
end subroutine find_pknl
! ***************************************************** !
!
subroutine calc_halo_pk(zred)
!
! ***************************************************** !
  use data,only : pi, ikmax, ik_max, ak, pk, ik_max_nl, ak_nl, pk_nl, rknl
  implicit none
  
  double precision, intent(in) :: zred
  integer ik, icount
  double precision growth
  double precision xlogr1, xlogr2, rmid, sig, d1, d2
  double precision diff, rneff, rncur, Del_lin, Del_nl
  double precision pq, ph
  external growth
! -----------------------------------------  

! calculate nonlinear wavenumber (rknl), effective spectral index (rneff) and 
! curvature (rncur) of the power spectrum at the desired redshift, using method 
! described in Smith et al (2002).

  icount = 0
!!!      xlogr1=-2.0    A.Taruya 2008/12/03
  xlogr1 = -3.d0
  xlogr2 = 3.5d0
10 rmid = (xlogr2+xlogr1)/2.d0
  rmid = 10**rmid

!!!!  A.Taruya 2008/12/03  force to stop iteration
  icount = icount + 1
  if(icount > 30) then
!!!!         write(6,*) ' computation of rknl fails to converge '
     rknl = 1.d0/rmid
     rneff = -3-d1
     rncur = -d2
     goto 20
  endif
!!!!  A.Taruya 2008/12/03 

  call wint(rmid,sig,d1,d2,zred)
  diff = sig-1.0
  if (abs(diff) <= 1d-3) then
     rknl = 1./rmid
     rneff = -3-d1
     rncur = -d2                  
  elseif (diff > 1d-3) then
     xlogr1 = log10(rmid)
     goto 10 
  elseif (diff < -1d-3) then
     xlogr2 = log10(rmid)
     goto 10
  endif

20 ik_max_nl = ik_max

  do ik = 1, ik_max_nl

! linear power spectrum !! Remeber => plin = k^3 * P(k) * constant
! constant = 4*pi*V/(2*pi)^3 

     Del_lin = growth(zred)**2 * ak(ik)**3 * pk(ik)  &
          &   / (2.d0*pi*pi)

! calculate nonlinear power according to halofit: pnl = pq + ph,
! where pq represents the quasi-linear (halo-halo) power and 
! where ph is represents the self-correlation halo term. 
 
     call halofit(zred,ak(ik),rneff,rncur,rknl, &
          &  Del_lin,Del_nl,pq,ph)
     
     ak_nl(ik) = ak(ik)
     pk_nl(ik) = Del_nl / Del_lin * pk(ik) * growth(zred)**2
     
  enddo
  return
end subroutine calc_halo_pk

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

! The subroutine wint, finds the effective spectral quantities
! rknl, rneff & rncur. This it does by calculating the radius of 
! the Gaussian filter at which the variance is unity = rknl.
! rneff is defined as the first derivative of the variance, calculated 
! at the nonlinear wavenumber and similarly the rncur is the second
! derivative at the nonlinear wavenumber. 

subroutine wint(r,sig,d1,d2,zred)
  
  implicit none
  double precision, parameter :: pi = 3.14159265358979d0
  double precision, intent(in) :: r, zred
  double precision, intent(out) :: sig, d1, d2
  double precision sum1, sum2, sum3, t, y, x, w1, w2, w3
  double precision rk, pklin, growth
  integer i, nint
  external growth

  nint = 3000
  sum1 = 0.d0
  sum2 = 0.d0
  sum3 = 0.d0
  do i = 1, nint
     t = (dble(i)-0.5d0)/dble(nint)
     y = -1.d0 + 1.d0/t
     rk = y
     call find_pk(rk, pklin)
     d2 = growth(zred)**2 * pklin * rk**3 / (2.d0*pi*pi)
     x = y*r
     w1 = exp(-x*x)
     w2 = 2*x*x*w1
     w3 = 4*x*x*(1-x*x)*w1
     sum1 = sum1 + w1*d2/y/t/t
     sum2 = sum2 + w2*d2/y/t/t
     sum3 = sum3 + w3*d2/y/t/t
  enddo
  sum1 = sum1/dble(nint)
  sum2 = sum2/dble(nint)
  sum3 = sum3/dble(nint)
  sig = sqrt(sum1)
  d1 = -sum2/sum1
  d2 = -sum2*sum2/sum1/sum1 - sum3/sum1
  
  return
end subroutine wint
      
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!
! halo model nonlinear fitting formula as described in 
! Appendix C of Smith et al. (2002)
!
subroutine halofit(zred,rk,rn,rncur,rknl,plin,pnl,pq,ph)
  use data,only : Omega_m, Omega_v, w_de, sigma8 
  implicit none
  
  double precision, intent(in) :: zred, rk, rn, rncur, rknl, plin
  double precision, intent(out) :: pnl, pq, ph
  double precision gam, a, amod, b, c, xmu, xnu, alpha, beta
  double precision f1, f2, f3, f4
  double precision om_m, om_v, omg_m, omg_v, aexp
  double precision y, p_cdm
  double precision f1a, f2a, f3a, f4a, f1b, f2b, f3b, f4b, frac
  external omg_m, omg_v
! --------------------------------------------  !
  
  aexp = 1.d0 / (1.d0 + zred)
  om_m = omg_m(aexp, Omega_m, Omega_v, w_de) 
  om_v = omg_v(aexp, Omega_m, Omega_v, w_de) 

  gam = 0.86485 + 0.2989*rn + 0.1631*rncur
  a = 1.4861 + 1.83693*rn + 1.67618*rn*rn + 0.7940*rn*rn*rn &
       & + 0.1670756*rn*rn*rn*rn - 0.620695*rncur
  a = 10**a      
  b = 10**(0.9463+0.9466*rn+0.3084*rn*rn-0.940*rncur)
  c = 10**(-0.2807+0.6669*rn+0.3214*rn*rn-0.0793*rncur)
  xmu = 10**(-3.54419+0.19086*rn)
  xnu = 10**(0.95897+1.2857*rn)
  alpha = 1.38848 + 0.3701*rn - 0.1452*rn*rn
  beta = 0.8291 + 0.9854*rn + 0.3400*rn**2
  
  if(abs(1-om_m) > 1d-2) then ! omega evolution 
     f1a = om_m**(-0.0732d0)
     f2a = om_m**(-0.1423d0)
     f3a = om_m**(0.0725d0)
     f1b = om_m**(-0.0307d0)
     f2b = om_m**(-0.0585d0)
     f3b = om_m**(0.0743d0)       
     frac = om_v/(1.d0-om_m) 
     f1 = frac*f1b + (1-frac)*f1a
     f2 = frac*f2b + (1-frac)*f2a
     f3 = frac*f3b + (1-frac)*f3a
  else         
     f1 = 1.d0
     f2 = 1.d0
     f3 = 1.d0
  endif

  y = (rk/rknl)
  
  ph = a*y**(f1*3)/(1+b*y**(f2)+(f3*c*y)**(3-gam))
  ph = ph/(1+xmu*y**(-1)+xnu*y**(-2))
  pq = plin*(1+plin)**beta/(1+plin*alpha)*exp(-y/4.0-y**2/8.0)

  pnl = pq + ph

  return
end subroutine halofit
 
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!
! evolution of omega matter with expansion factor
function omg_m(aa, om_m0, om_v0, w_de)
  implicit none
  double precision, intent(in) :: aa, om_m0, om_v0, w_de
  double precision omg_m, omega_k

  omega_k = om_m0 + om_v0 - 1.d0 
  omg_m = om_m0 / ( om_m0 + om_v0/aa**(3.0*w_de) - omega_k*aa )
  return
end function omg_m
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!
! evolution of omega lambda with expansion factor
function omg_v(aa, om_m0, om_v0, w_de)      
  implicit none
  double precision, intent(in) :: aa, om_m0, om_v0, w_de
  double precision omg_v, omega_k

  omega_k = om_m0 + om_v0 - 1.d0 
  omg_v = om_v0/aa*(3.0*w_de)  & 
  &     / ( om_m0 + om_v0/aa**(3.0*w_de) - omega_k*aa )
  return
end function omg_v
