! ******************************************************* !
!                                                         !
!      A Closure theory for non-linear evolution of       !
!      Power spectrum                                     !
!                                                         !
!            Time-stamp: <2011-04-14 13:17:35 t-okamura>  !
! ******************************************************* !
module data
  implicit none

!!! constants
  ! fundamental constants
  double precision, parameter :: pi = 3.14159265358979d0

  ! cosmological parameters
  double precision Omega_m, Omega_v, w_de, sigma8
  
!!! input & output datas, and input parameters for calculation
  ! power spectrum data
  integer, parameter :: ikmax = 5000
  integer ik_max
  double precision, dimension(ikmax) :: ak, pk
  double precision, dimension(ikmax) :: pk_A11, pk_A12, pk_A22
  double precision, dimension(ikmax) :: pk_B11, pk_B12, pk_B22
  double precision, dimension(ikmax) :: pk_2loop11, pk_2loop12, pk_2loop22
  ! wave number
  integer ibox
  double precision kmin, kmax, Lbox, k_cut_min, k_cut_max
  ! alpha correction
  integer ialpha
  ! redshift
  double precision zred
  ! others 
  integer itype, ijob, ik_init, iloop
  ! file name
  character*100 fout
  character*100, parameter :: fout_RPT_1loop = 'renormalized_pk2_RPT_1loop.dat'
  character*100, parameter :: fout_CT_1loop = 'renormalized_pk2_Closure_1loop.dat'
  character*100, parameter :: fout_RPT_2loop = 'renormalized_pk2_RPT_2loop.dat'
  character*100, parameter :: fout_CT_2loop = 'renormalized_pk2_Closure_2loop.dat'

!!! temporary data
  ! redshift and growth rate 
  double precision, parameter :: zinit = 500.d0
  double precision eeta, eeta0, eta, eta0

  ! 1-loop propagator
  double precision, dimension(ikmax) :: ff, gg, hh, ii

  ! halo model power spectrum
  integer ik_max_nl
  double precision rknl
  double precision, dimension(ikmax) :: ak_nl, pk_nl

  ! alpha correction
  integer, parameter :: izmax = 100
  integer iz_max
  double precision, dimension(izmax) :: etaa, alphaz
  
!!! MPI
  integer :: isize = 1
  integer :: irank = 0
  
end module data

program pk_closure
  use data
  implicit none
  integer ik
! -------------------------------------------------
  
  write(6,*)
  write(6,*) '*****************************************************'
  write(6,*) '  Improved PT calculation for matter power spectrum  '
  write(6,*) '*****************************************************'
  write(6,*)
  write(6,*) 'input output redshift' 
  read(5,*) zred

  write(6,*) 'choose type of propagator: PRT[0], Closure theory[1]'
  read(5,*) itype
  if((itype /= 0) .and. (itype /=1) ) stop

  write(6,*) 'choose loop order: 1-loop [1] or 2-loop [2]'
  read(5,*)iloop
  if((iloop /= 1) .and. (iloop /=2) ) stop
  if(itype == 0 .and. iloop == 1) fout = fout_RPT_1loop
  if(itype == 0 .and. iloop == 2) fout = fout_RPT_2loop
  if(itype == 1 .and. iloop == 1) fout = fout_CT_1loop
  if(itype == 1 .and. iloop == 2) fout = fout_CT_2loop

  write(6,*) 'including the correction in propagator? :y[0],n[1]'
  read(5,*) ialpha
  if((ialpha /= 0) .and. (ialpha /= 1)) stop

  write(6,*) 'restart suspended run[0], start a new calcuation[1]'
  read(5,*) ijob
  if(ijob == 0) then
     call search(fout,ik_init)
  elseif(ijob == 1)then
     ik_init = 1
  else 
     stop
  endif

  write(6,*) 'including finite-volume effect? :type (0 or 1, L_box)'
  write(6,*) ' (e.g., (0, 300) for boxsize L_box=300Mpc/h ) '
  write(6,*) ' (      (1, 100) for ignoring finite-volume effect) '
  read(5,*) ibox, Lbox
  if((ibox /= 0) .and. (ibox /= 1)) stop

  !    //////// Load transfer function ////////  
  call load_transfer_function
  write(6,*) ' load (linear) matter power spectrum, done '
  
  !     //////// Set cosmological parameters ////////
  call set_cosmological_params
  
  !     //////// Sigma8 normalization  ////////
  call normalization_trapez
  write(6,*) ' sigma8 normalization, done '
  
  !    //////// Get linear growth rate ////////
  call get_linear_growth
  
  !     //////// Alpha correction in propagator ////////
  if(ialpha == 0) then 
     call calc_alpha_correction(zred, ibox, Lbox)
     write(6,*) ' computing alpha(z), done '
  endif
  
  !     //////// Truncation of low-k, high-k  in linear P(k) ////////
  kmin = 5d-4 ! default value
  kmax = 5.d0 ! default value
  call truncation_k
  
  !     //////// 1-loop propagator ////////
  call one_loop_propagator
  write(6,*) ' computing 1-loop propagator, done '
     
 
!     //////// Renormalized P(k)  ////////
  write(6,*) ' input k_min for output data'
  read(5,*) k_cut_min
  write(6,*) ' input k_max for output data'
  read(5,*) k_cut_max
  
  if(iloop == 1) call calc_renormalized_pk_1loop
  if(iloop == 2) call calc_renormalized_pk_2loop
  write(6,*) ' computing renormalized P(k), done '
  
!     //////// output the results  //////// 
  if(ijob == 0) open(10,file=fout,status='unknown', access='append')
  if(ijob == 1) open(10,file=fout,status='unknown', access='sequential')
  
  if(iloop == 1) then
     do ik = ik_init, ik_max
        write(10,'(1p8e18.10)') ak(ik), eeta**2*pk(ik), pk_A11(ik), &
             &  pk_A12(ik), pk_A22(ik), pk_B11(ik), pk_B12(ik), pk_B22(ik)
     enddo
  elseif(iloop == 2) then
     do ik = ik_init, ik_max
        write(10,'(1p5e18.10)') ak(ik), eeta**2*pk(ik), pk_2loop11(ik), &
             &                 pk_2loop12(ik), pk_2loop22(ik)
     enddo
  endif

  close(10)
end program pk_closure
! ******************************************************* !

subroutine get_linear_growth
  
! ******************************************************* !
  use data,only : zred, zinit, eeta, eeta0, eta, eta0
  implicit none
  double precision growth
  external growth

  eeta = growth(zred)
  eeta0 = growth(zinit)

  eta = log(eeta)
  eta0 = log(eeta0)
  return
end subroutine get_linear_growth
! ******************************************************* !

subroutine load_transfer_function
  
! ******************************************************* !
  
  !     input file is assumed to be the transfer function data 
  !     created by CAMB code. that is, the data sorting is assumed
  !     to be 
  !
  !     k [h/Mpc],  P(k) [Mpc^3/h^3]

  use data, only : ak, pk, ikmax, ik_max
  implicit none
  
  integer ik
  character infile*50
!   --------------------------------------------------------

  write(6, *) ' Type file name of matter P(k) data'
  read(5, *) infile  
  open(9, file=infile, status='old', action='read')
  
  do ik = 1, ikmax
     read(9,*,END=10) ak(iK), pk(ik)   
  enddo
  
10 continue
  ik_max = ik - 1
  
  write(6,*) 'ak(1)=', ak(1)
  write(6,*) 'ak(ik_max)=', ak(ik_max)
  write(6,*) 'ik_max=', ik_max
  
  close(9)
  
end subroutine load_transfer_function
! ******************************************************* !
!
subroutine set_cosmological_params
!
! ******************************************************* !
!
!     Set cosmological parameters for linear growth factor,
!     growth-rate parameter, and sigma8 normalization 
!
!     Note.--
!     assume a flat cosmological model (Omega_v = 1-Omega_m)
!
  use data,only : Omega_m, Omega_v, w_de, sigma8
  implicit none
  integer iparams
!  ---------------------------------------------------

  write(6,*)
  write(6,*) '*** Set cosmological parameters ***' 
  write(6,*)

!     Default cosmological parameters (you can change them appropriately)
  Omega_m = 0.279d0
  Omega_v = 1.d0 - Omega_m
  w_de = -1.d0
  sigma8 = 0.817d0
!
5 write(6,*) '[1] omega_m =',Omega_m
  write(6,*) '[2]       w =',w_de
  write(6,*) '[3]  sigma8 =',sigma8
  write(6,*) 
  write(6,*) 'Note--. spatial curvature is forced to be flat'
  write(6,*) '        i.e., omega_m = 1 - omega_DE           '

  write(6,*)
  write(6,*) 'change cosmological parameter? [1-3] or n[0]'
  read(5,*)  iparams
  if(iparams == 0) goto 8
  if(iparams == 1) then
     write(6,*) 'type omega_m'
     read(5,*) Omega_m
     Omega_v = 1.d0 - Omega_m
  elseif(iparams == 2) then
     write(6,*) 'type w'
     read(5,*) w_de
  elseif(iparams == 3) then
     write(6,*) 'type sigma8'
     read(5,*) sigma8
  endif
  goto 5

8 continue

end subroutine set_cosmological_params
! ******************************************************* !
!
subroutine normalization_trapez

! ******************************************************* !
!  normalization of power spectrum amplitude by sigma8
!  using a simple trapezoidal integration
  use data, only : pi, ak, pk, ik_max, sigma8
  implicit none
  
  double precision, parameter :: r_th = 8.d0
  integer ik
  double precision  const, x, x2
  double precision  W_TH, sigma_a, sigma_b

! ---------------------------------------------------

  x = ak(1) * r_th
  x2 = x * x
  if(x < 1d-3) then
     W_TH = 1.d0 - 0.1d0 * x2  + x2*x2 / 280.d0 
  else
     W_TH = 3.d0 * (sin(x) - x * cos(x))/x2/x
  endif
  sigma_a = W_TH * W_TH * pk(1) * ak(1) * ak(1)
  sigma_a = sigma_a / (2.d0*pi*pi)

  const = 0.d0 
  do ik = 2, ik_max
     x = ak(ik) * r_th
     x2 = x * x
     if(x < 1d-3) then
        W_TH = 1.d0 - 0.1d0 * x2  + x2*x2 / 280.d0 
     else
        W_TH = 3.d0 * (sin(x) - x * cos(x))/x2/x
     endif
     sigma_b = W_TH * W_TH * pk(ik) * ak(ik) * ak(ik) 
     sigma_b = sigma_b / (2.d0*pi*pi)
     const = const   &
          & + (sigma_a + sigma_b) * ( ak(ik) - ak(ik-1) ) * 0.5d0
     sigma_a = sigma_b
  enddo

  do ik = 1, ik_max
     pk(ik) = sigma8 * sigma8 / const * pk(ik) 
  enddo
  
end subroutine normalization_trapez
    
! ******************************************************* !

subroutine truncation_k

! ******************************************************* !
  use data,only : ikmax, ik_max, ak, pk, pi, kmin, kmax, &
       & ibox, Lbox 
  implicit none

  integer ik, ikk
  double precision akk(ikmax), pkk(ikmax)

! -----------------------------------------------------

  if(ibox == 0) kmin = 2.d0 * pi / Lbox

  akk(1:ik_max) = ak(1:ik_max)
  pkk(1:ik_max) = pk(1:ik_max)

  ikk = 1
  do ik = 1, ik_max
     if( (akk(ik) >= kmin) .and. (akk(ik) <= kmax) &
          & .and. (mod(ik,2) == 0) ) then
        ak(ikk) = akk(ik)
        pk(ikk) = pkk(ik)
        ikk = ikk + 1
     endif
  enddo

  ik_max = ikk -1
  
  write(6,*) 'ak(1)=', ak(1)
  write(6,*) 'ak(ik_max)=', ak(ik_max)
  write(6,*) 'ik_max=', ik_max
  
end subroutine truncation_k

  
! ******************************************************* !
!
subroutine search(fname,ik_init)
!
! ******************************************************* !
!
!     search the number record of last output data
!
  use data,only : ikmax
  implicit none
  character*100, intent(in) :: fname
  integer, intent(out) :: ik_init
  integer ik
  double precision ak_dummy, dummy
! ------------------------------------------

  open(10,file=fname,status='unknown')

  do ik = 1, ikmax
     read(10,*,END=10) ak_dummy, dummy, dummy
  enddo

10 ik_init = ik

  write(6,*) ik_init, ak_dummy

  close(10)
end subroutine search
! ******************************************************* !

subroutine find_pk(kk, pklin)

! ******************************************************* !
  use data, only : ik_max, ak, pk
  implicit none
  double precision, intent(in) :: kk
  double precision, intent(out) :: pklin
  integer j, jmin, jmax
  double precision s, ds
 
! -------------------------------------------
  
  if( kk <= ak(ik_max)) then
     call hunt(ak, ik_max, kk, j)
     
     jmin = max(1, j - 2)
     jmax = min(ik_max, j + 2)
  
     call polint(ak(jmin:jmax),pk(jmin:jmax),jmax-jmin+1,kk,s,ds)
    
  else
     call extrapolation(kk,s)
  endif
  
  pklin = s
  return
end subroutine find_pk
! ******************************************************* !
!
subroutine extrapolation(kk, pklin)
!
! ******************************************************* !
!
!     Extrapolation of linear P(k) data 
!
!     The extrapolation is done assuming the single power-law  
!     form of P(k) on small-scales. The slope of power-law 
!     function is determined by using the last 15 points of 
!     P(k) data.  
!
  use data, only : ikmax, ik_max, ak, pk
  implicit none
  integer, parameter :: ikdelta = 15
  double precision, intent(in) :: kk
  double precision, intent(out) :: pklin
  integer ik
  double precision dlnp_dlnk, n_eff
! --------------------------------------

  dlnp_dlnk = 0.d0

  do ik = ik_max - ikdelta, ik_max-1
     dlnp_dlnk = dlnp_dlnk &
          & + log(pk(ik+1)/pk(ik-1))/log(ak(ik+1)/ak(ik-1))
  enddo
  
  n_eff = dlnp_dlnk / dble(ikdelta)
  if(n_eff < -3.d0) n_eff = -3.d0
  if(kk > 1d3) n_eff = -3.d0

  pklin = pk(ik_max) * ( kk / ak(ik_max) ) ** n_eff
  return
end subroutine extrapolation

! ******************************************************* !

function growth(zred)
!
! ******************************************************* !
!
!    Linear growth factor for flat cosmology
!  
  use data,only : Omega_m, Omega_v, w_de
  implicit none
  double precision, intent(in) :: zred
  double precision growth, a, b, c
  double precision zred1, zi, zz, gz, g0
! --------------------------------------------

  a = -1.d0 / (3.d0 * w_de)
  b = (w_de - 1.d0)/ (2.d0 * w_de)
  c = 1.d0 - 5.d0 / (6.d0 * w_de)

  zred1 = 1.d0 + zred
  zi = - Omega_v / Omega_m
  zz = zi * zred1**(3.d0*w_de) 

  call HYGFX(a,b,c,zi, g0)  
  call HYGFX(a,b,c,zz, gz)  

  growth = (gz/g0) / zred1
  return
end function growth
! ************************************************************

subroutine polint(xa,ya,n,x,y,dy)
  
!  ************************************************************
  
  implicit none
  integer, parameter :: NMAX = 10
  integer, intent(in) :: n
  double precision, intent(in) :: x, xa(n), ya(n)
  double precision, intent(out) :: y, dy
  integer i, m, ns
  double precision den, dif, dift, ho, hp, w, c(NMAX), d(NMAX)

  ns = 1
  dif = abs(x-xa(1))
  do i = 1, n
     dift = abs(x-xa(i))
     if (dift < dif) then
        ns = i
        dif = dift
     endif
     c(i) = ya(i)
     d(i) = ya(i)
  enddo
  
  y = ya(ns)
  ns = ns - 1
  do m = 1, n-1
     do i = 1, n-m
        ho = xa(i) - x
        hp = xa(i+m) - x
        w = c(i+1) - d(i)
        den = ho - hp
        if(den == 0.d0)stop 'failure in polint'
        den = w/den
        d(i) = hp*den
        c(i) = ho*den
     enddo
     if (2*ns < n-m)then
        dy = c(ns+1)
     else
        dy = d(ns)
        ns = ns-1
     endif
     y = y + dy
  enddo
  return
end subroutine polint

! ************************************************ !
        
subroutine hunt(xx,n,x,jlo)

! ************************************************ !

  implicit none
  integer, intent(in) :: n
  double precision, intent(in) :: xx(n), x
  integer, intent(out) :: jlo
  integer inc, jhi, jm
  logical ascnd

  ascnd = xx(n).gt.xx(1)
  if(jlo <= 0 .or. jlo > n)then
     jlo = 0
     jhi = n + 1
     goto 3
  endif
  inc = 1
  if(x >= xx(jlo).eqv.ascnd)then
1    jhi = jlo + inc
     if(jhi > n)then
        jhi = n + 1
     else if(x >= xx(jhi).eqv.ascnd)then
        jlo = jhi
        inc = inc + inc
        goto 1
     endif
  else
     jhi = jlo
2    jlo = jhi - inc
     if(jlo < 1)then
        jlo = 0
     else if(x < xx(jlo).eqv.ascnd)then
        jhi = jlo
        inc = inc + inc
        goto 2
     endif
  endif
3 if(jhi-jlo == 1)return
  jm = (jhi+jlo)/2
  if(x > xx(jm).eqv.ascnd)then
     jlo = jm
  else
     jhi = jm
  endif
  goto 3
end subroutine hunt
! ******************************************************* !

subroutine HYGFX(A,B,C,X,HF)

! ******************************************************* !
!
!       ====================================================
!       Purpose: Compute hypergeometric function F(a,b,c,x)
!       Input :  a --- Parameter
!                b --- Parameter
!                c --- Parameter, c <> 0,-1,-2,...
!                x --- Argument   ( x < 1 )
!       Output:  HF --- F(a,b,c,x)
!       Routines called:
!            (1) GAMMAX for computing gamma function
!            (2) PSI for computing psi function
!       ====================================================
!
  IMPLICIT DOUBLE PRECISION (A-H,O-Z)
  logical L0,L1,L2,L3,L4,L5
  pi = 3.141592653589793d0
  EL = 0.5772156649015329d0

  L0 = (C == int(C)).and.(C < 0.d0)
  L1 = (1.d0-X < 1.d-15).and.(C-A-B <= 0.d0)
  L2 = (A == int(A)).and.(A < 0.d0)
  L3 = (B == int(B)).and.(B < 0.d0)
  L4 = (C-A == int(C-A)).and.(C-A <= 0.d0)
  L5 = (C-B == int(C-B)).and.(C-B <= 0.d0)
  if (L0.or.L1) then
     write(*,*)'The hypergeometric series is divergent'
     return
  endif
  EPS = 1.d-15
  if (X > 0.95d0) EPS = 1.d-8
  if ((X == 0.d0).or.(A == 0.d0).or.(B == 0.d0)) then
     HF = 1.d0
     return
  elseif( (1.d0-X == EPS).and.(C-A-B > 0.d0) ) then
     call GAMMAX(C,GC)
     call GAMMAX(C-A-B,GCAB)
     call GAMMAX(C-A,GCA)
     call GAMMAX(C-B,GCB)
     HF = GC*GCAB/(GCA*GCB)
     return
  elseif( (1.d0+X <= EPS).and.(abs(C-A+B-1.d0) <= EPS) ) then
     G0 = sqrt(PI)*2.d0**(-A)
     call GAMMAX(C,G1)
     call GAMMAX(1.0D0+A/2.0-B,G2)
     call GAMMAX(0.5D0+0.5*A,G3)
     HF = G0*G1/(G2*G3)
     return
  elseif (L2.or.L3) then
     if (L2) NM = int(abs(A))
     if (L3) NM = int(abs(B))
     HF = 1.d0
     R = 1.d0
     do K = 1, NM
        R = R*(A+K-1.d0)*(B+K-1.d0)/(K*(C+K-1.d0))*X
        HF = HF + R
     enddo
     return
  elseif (L4.or.L5) then
     if (L4) NM = int(abs(C-A))
     if (L5) NM = int(abs(C-B))
     HF = 1.d0
     R = 1.0d0
     do K = 1, NM
        R = R*(C-A+K-1.d0)*(C-B+K-1.d0)/(K*(C+K-1.d0))*X
        HF = HF + R
     enddo
     HF = (1.d0-X)**(C-A-B)*HF
     return
  endif
  AA = A
  BB = B
  X1 = X
  if (X < 0.d0) then
     X=X/(X-1.d0)
     if ( (C > A).and.(B < A).and.(B > 0.d0) ) then
        A = BB
        B = AA
     endif
     B = C - B
  endif
  if (X >= 0.75d0) then
     GM = 0.d0
     if (abs(C-A-B-int(C-A-B)) < 1d-15) then
        M = int(C-A-B)
        call GAMMAX(A,GA)
        call GAMMAX(B,GB)
        call GAMMAX(C,GC)
        call GAMMAX(A+M,GAM)
        call GAMMAX(B+M,GBM)
        call PSI(A,PA)
        call PSI(B,PB)
        if (M /= 0) GM = 1.d0
        do J = 1,abs(M)-1
           GM = GM*J
        enddo
        RM = 1.d0
        do J = 1, abs(M)
           RM = RM*J
        enddo
        F0 = 1.d0
        R0 = 1.d0
        R1 = 1.d0
        SP0 = 0.d0
        SP = 0.d0
        if (M >= 0) then
           C0 = GM*GC/(GAM*GBM)
           C1 = -GC*(X-1.d0)**M/(GA*GB*RM)
           do K = 1, M-1
              R0 = R0*(A+K-1.d0)*(B+K-1.d0)/(K*(K-M))*(1.d0-X)
              F0 = F0 + R0
           enddo
           do K = 1, M
              SP0 = SP0+1.d0/(A+K-1.d0)+1.d0/(B+K-1.d0)-1.d0/K
           enddo
           F1 = PA+PB+SP0+2.d0*EL+log(1.d0-X)
           do K = 1, 250
              SP = SP+(1.d0-A)/(K*(A+K-1.d0))+(1.d0-B)/(K*(B+K-1.d0))
              SM = 0.d0
              do J = 1, M
                 SM = SM+(1.d0-A)/((J+K)*(A+J+K-1.d0))+1.d0/  &
                 &                    (B+J+K-1.d0)
              enddo
              RP = PA+PB+2.d0*EL+SP+SM+log(1.d0-X)
              R1 = R1*(A+M+K-1.d0)*(B+M+K-1.d0)/(K*(M+K))*(1.d0-X)
              F1 = F1 + R1*RP
              if (abs(F1-HW) < abs(F1)*EPS) goto 60
              HW = F1
           enddo
60         HF = F0*C0 + F1*C1
        elseif (M < 0) then
           M = -M
           C0 = GM*GC/(GA*GB*(1.d0-X)**M)
           C1 = -(-1)**M*GC/(GAM*GBM*RM)
           do K = 1, M-1
              R0 = R0*(A-M+K-1.0d0)*(B-M+K-1.d0)/(K*(K-M))*(1.d0-X)
              F0 = F0 + R0
           enddo
           do K = 1, M
              SP0 = SP0+1.d0/K
           enddo
           F1 = PA + PB - SP0 +2.d0*EL + log(1.d0-X)
           do K = 1, 250
              SP = SP + (1.d0-A)/(K*(A+K-1.d0))+(1.d0-B)/(K*(B+K-1.d0))
              SM = 0.d0
              do J = 1, M
                 SM= SM + 1.d0/(J+K)
              enddo
              RP = PA + PB + 2.d0*EL + SP - SM + log(1.d0-X)
              R1 = R1*(A+K-1.d0)*(B+K-1.d0)/(K*(M+K))*(1.d0-X)
              F1 = F1 + R1*RP
              if (abs(F1-HW) < abs(F1)*EPS) goto 85
              HW = F1
           enddo
85         HF = F0*C0 + F1*C1
        endif
     else
        call GAMMAX(A,GA)
        call GAMMAX(B,GB)
        call GAMMAX(C,GC)
        call GAMMAX(C-A,GCA)
        call GAMMAX(C-B,GCB)
        call GAMMAX(C-A-B,GCAB)
        call GAMMAX(A+B-C,GABC)
        C0 = GC*GCAB/(GCA*GCB)
        C1 = GC*GABC/(GA*GB)*(1.d0-X)**(C-A-B)
        HF = 0.d0
        R0 = C0
        R1 = C1
        do K = 1, 250
           R0 = R0*(A+K-1.d0)*(B+K-1.d0)/(K*(A+B-C+K))*(1.d0-X)
           R1 = R1*(C-A+K-1.d0)*(C-B+K-1.d0)/(K*(C-A-B+K))*(1.d0-X)
           HF = HF + R0 + R1
           if (abs(HF-HW) < abs(HF)*EPS) goto 95
           HW = HF
        enddo
95      HF = HF + C0 + C1
     endif
  else
     A0 = 1.d0
     if ( (C > A).and.(C < 2.d0*A).and.(C > B).and.(C < 2.d0*B) ) then
        A0 = (1.d0-X)**(C-A-B)
        A = C - A
        B = C - B
     endif
     HF = 1.d0
     R = 1.d0
     do K = 1, 250
        R = R*(A+K-1.d0)*(B+K-1.d0)/(K*(C+K-1.d0))*X
        HF = HF + R
        if (abs(HF-HW) <= abs(HF)*EPS) goto 105
        HW = HF
     enddo
105  HF = A0*HF
  endif
  if (X1 < 0.d0) then
     X = X1
     C0 = 1.d0/(1.d0-X)**AA
     HF = C0*HF
  endif
  A = AA
  B = BB
  if (K > 120) write(*,115)
115 format(1X,'Warning! You should check the accuracy')
  return
end subroutine HYGFX

! ******************************************************* !

subroutine GAMMAX(X,GA)

! ******************************************************* !
!
!       ==================================================
!       Purpose: Compute gamma function â(x)
!       Input :  x  --- Argument of â(x)
!                       ( x is not equal to 0,-1,-2,úúú)
!       Output:  GA --- â(x)
!       ==================================================
!
  implicit double precision (A-H,O-Z)
  dimension G(26)
  PI = 3.141592653589793d0
  if (X == int(X)) then
     if (X > 0.d0) then
        GA = 1.d0
        M1 = X - 1
        do K = 2, M1
           GA = GA*K
        enddo
     else
        GA = 1d+300
     endif
  else
     if (abs(X) > 1.d0) then
        Z = abs(X)
        M = int(Z)
        R = 1.d0
        do K = 1, M
           R = R*(Z-K)
        enddo
        Z = Z - M
     else
        Z = X 
     endif
     DATA G/1.0D0,0.5772156649015329d0, &
     &          -0.6558780715202538d0, -0.420026350340952d-1, &
     &          0.1665386113822915d0,-.421977345555443d-1, &
     &          -.96219715278770d-2, .72189432466630d-2, &
     &          -.11651675918591d-2, -.2152416741149d-3, &
     &          .1280502823882d-3, -.201348547807d-4, &
     &          -.12504934821d-5, .11330272320d-5, &
     &          -.2056338417d-6, .61160950d-8, &
     &          .50020075d-8, -.11812746d-8, &
     &          .1043427d-9, .77823d-11, &
     &          -.36968d-11, .51d-12, &
     &          -.206d-13, -.54d-14, .14d-14, .1d-15/
     GR = G(26)
     do K = 25, 1, -1
        GR = GR*Z + G(K)
     enddo
     GA = 1.d0/(GR*Z)
     if (abs(X) > 1.d0) then
        GA = GA*R
        if (X < 0.d0) GA = -PI/(X*GA*sin(PI*X))
     endif
  endif
  return
end subroutine GAMMAX

! ******************************************************* !

subroutine PSI(X,PS)

! ******************************************************* !
!
!       ======================================
!       Purpose: Compute Psi function
!       Input :  x  --- Argument of psi(x)
!       Output:  PS --- psi(x)
!       ======================================
!
  implicit double precision (A-H,O-Z)
  XA = abs(X)
  pi = 3.141592653589793d0
  EL = 0.5772156649015329d0
  S = 0.d0
  if ( (X == int(X)).and.(X <= 0.d0) ) then
     PS = 1d+300
     return
  elseif (XA == int(XA)) then
     N = XA
     do K = 1, N-1
        S = S + 1.d0/K
     enddo
     PS = -EL + S
  elseif (XA+0.5 == int(XA+.5)) then
     N = XA - 0.5
     do  K = 1, N
        S = S + 1.d0/(2.d0*K-1.d0)
     enddo
     PS = -EL + 2.0d0*S -1.386294361119891d0
  else
     if (XA < 10.d0) then
        N = 10 - int(XA)
        do K = 0, N-1
           S = S + 1.d0/(XA+K)
        enddo
        XA = XA + N
     endif
     X2 = 1.d0/(XA*XA)
     A1 = -0.8333333333333d-01
     A2 = 0.83333333333333333d-02
     A3 = -0.39682539682539683d-02
     A4 = 0.41666666666666667d-02
     A5 = -0.75757575757575758d-02
     A6 = 0.21092796092796093d-01
     A7 = -0.83333333333333333d-01
     A8 = 0.4432598039215686d0
     PS = log(XA) - 0.5d0/XA + X2*(((((((A8*X2+A7)*X2+ &
          &        A6)*X2+A5)*X2+A4)*X2+A3)*X2+A2)*X2+A1)
     PS = PS - S
  endif
  if (X < 0.d0) PS = PS - PI*cos(PI*X)/sin(PI*X)-1.d0/X
  return
end subroutine PSI
