! ******************************************************* !
!
subroutine calc_renormalized_pk_1loop
!
! ******************************************************* !

!     computing renormalized power spectra up to 1-loop order 
!     using gaussian quadrature method
  use data,only : pi, ik_init, ik_max, eta, eta0, eeta0, ak, pk, &
       & pk_A11, pk_A12, pk_A22, pk_B11, pk_B12, pk_B22, &
       & irank, isize, itype, kmin, kmax
  implicit none

  integer, parameter :: ixmax = 400
  integer :: flag = 0
  integer ik, ix
  double precision G11, G12, G21, G22, pklin_kx
  double precision k, xmin, xmax, x, fact, mumin, mumax
  double precision intB11, intB12, intB22
  double precision wx(ixmax), xx(ixmax)
!    -------------------------------------------------
 
!     //////// calculation of A-part P(k) ////////
  pk_A11(:) = 0.d0
  pk_A12(:) = 0.d0
  pk_A22(:) = 0.d0

  do ik = ik_init, ik_max
     if(itype == 0) then 
        call G_ab_RPT(ak(ik), eta, eta0, G11, G12, G21, G22)
     elseif(itype == 1) then
        call G_ab_Closure(ak(ik), eta, eta0, G11, G12, G21, G22)
     endif
     fact = pk(ik) * eeta0**2
     pk_A11(ik) = (G11 + G12) * (G11 + G12) * fact
     pk_A12(ik) = (G11 + G12) * (G21 + G22) * fact
     pk_A22(ik) = (G21 + G22) * (G21 + G22) * fact
  enddo
!
!     //////// calculation of B-part P(k) ////////c
!
  kmin = ak(1)
  kmax = ak(ik_max)
  pk_B11(:) = 0.d0
  pk_B12(:) = 0.d0
  pk_B22(:) = 0.d0

  !$OMP PARALLEL DEFAULT(SHARED), &
  !$OMP & PRIVATE(ik, k, x, xmin, xmax, xx, wx, &
  !$OMP &   ix, pklin_kx, mumin, mumax, &
  !$OMP &   intB11, intB12, intB22, fact, flag)
  !$OMP DO SCHEDULE(DYNAMIC,1)
  do ik = ik_init + irank, ik_max, isize

     k = ak(ik)
     xmin = kmin / k
     xmax = kmax / k
         
!     skip the integration for negligible contribution from small-k 
     if( (k < 0.01d0) .or. (flag == 1) ) goto 99

!     integration by Gaussian quadratures method
     call gauleg(log(xmin), log(xmax), xx, wx, ixmax)
         
     do ix = 1, ixmax
        x = exp(xx(ix))
        call find_pk(k*x, pklin_kx)

        mumin = max(-1.d0, 0.5d0*(1.d0+x**2-xmax**2)/x)
        mumax = min( 1.d0, 0.5d0*(1.d0+x**2-xmin**2)/x)
        if(x >= 0.5d0) mumax= 0.5d0/x
        call int_pk_B(k, x, eta, eta0, itype, mumin, mumax, &
             & intB11, intB12, intB22) 

        pk_B11(ik) = pk_B11(ik) + wx(ix) * pklin_kx * intB11
        pk_B12(ik) = pk_B12(ik) + wx(ix) * pklin_kx * intB12
        pk_B22(ik) = pk_B22(ik) + wx(ix) * pklin_kx * intB22
     enddo
     
     fact = k**3 / (2.d0*pi**2) * eeta0**4 * 2.d0
     pk_B11(ik) = pk_B11(ik) * fact
     pk_B12(ik) = pk_B12(ik) * fact
     pk_B22(ik) = pk_B22(ik) * fact

!     criteria for terminating integration (2)
     if((ik > int(ik_max/2))  &
          & .and. (max(pk_A11(ik),pk_B11(ik)) <= 1d-20)) flag = 1

99   continue
  enddo
  !$OMP END DO
  !$OMP END PARALLEL

  return
end subroutine calc_renormalized_pk_1loop

! ******************************************************* !

subroutine int_pk_B(k, x, eta, eta0, itype, mumin, mumax, &
     & intB11, intB12, intB22)

! ******************************************************* !

!     integral over mu by Gaussian quadratures method

  implicit none
  integer, parameter :: imumax = 10
  integer, intent(in) :: itype
  double precision, intent(in) :: k, x, eta, eta0, mumin, mumax
  double precision, intent(out) :: intB11, intB12, intB22

  integer imu
  double precision pklin_mu, Lambda1, Lambda2, xabs2, fact
  double precision xmu(imumax), wmu(imumax)
  
  intB11 = 0.d0
  intB12 = 0.d0
  intB22 = 0.d0

  call gauleg(mumin, mumax, xmu, wmu, imumax)

  do imu = 1, imumax
     xabs2 = 1.d0 + x*x - 2.d0*xmu(imu)*x
     call find_pk(k*sqrt(xabs2), pklin_mu)
     call Lambda(itype, k, x, xmu(imu), xabs2, eta, eta0, Lambda1, Lambda2) 

     fact = wmu(imu) * x**2 * pklin_mu 
     intB11 = intB11 + Lambda1**2 * fact
     intB12 = intB12 + Lambda1*Lambda2 * fact
     intB22 = intB22 + Lambda2**2 *fact
  enddo
  
  intB11 = intB11 * x
  intB12 = intB12 * x
  intB22 = intB22 * x
  
  return
end subroutine int_pk_B
